#!/usr/bin/env python3
# -----------------------------------------------------------------------
# This is a script to download reference data
#
# To run this script, you need to:
#
#   Enter your ApiKey as parameter prior to run the script:
#       Example: download-reference-data.py --api-key=<yourAPIKey>
#   Please specify also the installer for your preferred platform:
#       Example: download-reference-data.py --api-key=<yourAPIKey> --installer=<installer>
#       See https://connect.uniserv.com?redirect=/onpremises/reference-data for the list of available installers.
#
# -----------------------------------------------------------------------
import argparse
import hashlib
import json
import os
import re
import secrets
import shutil
import sys
import urllib.request
from pathlib import Path
from urllib.error import HTTPError

ENDPOINT = 'https://download.prd.connect.uni-eng.de'
VERSION = 'v1.2.3'
exceptions = []
skippedDownloads = []
downloadedArtifacts = []
downloadOutputFileName = 'downloadedFiles.txt'
tableTemplate = '{:40} {:20} {:40}'


# Command line arguments
def get_commandline_arguments():
    parser = argparse.ArgumentParser(description='''Downloads reference data.
 The API key is mandatory to identify the Uniserv CONNECT account and its download permissions.''')

    parser.add_argument('--api-key', '-k', type=str, required=True,
                        help="API key used to access the reference data download resources.")
    parser.add_argument('--installer', '-i', type=str, required=False,
                        help="Please specify the installer name for your platform. For available installers see "
                             "https://connect.uniserv.com?redirect=/onpremises/reference-data. "
                             "If not specified, all installers will be downloaded.")
    parser.add_argument('--proxy', '-x', type=str, required=False,
                        help="The required IP address or the hostname of the proxy server, "
                             "optionally followed by a colon and the port number, "
                             "e.g. myproxy.company.de, myproxy.company.de:1080, 172.25.208.1:1080")
    parser.add_argument('--proxy-protocol', '-xp', type=str, required=False,
                        help="The protocol to use to connect to the proxy server: http or https")
    parser.add_argument('--proxy-user', '-xu', type=str, required=False,
                    help="The optional username to specify when proxy authentication is required.")
    parser.add_argument('--proxy-password', '-xpw', type=str, required=False,
                    help="The optional password to specify when proxy authentication is required.")
    parser.add_argument('--version', '-v',
                        action='version',
                        version=VERSION,
                        help="Show the current version of the script.")

    args = parser.parse_args()

    return args


class DownloadFailed(Exception):
    """Internal exception, used with a list of error messages as argument"""
    pass


class ErrorIgnore(object):
    def __init__(self, errors, error_return=None, error_call=None):
        self.errors = errors
        self.error_return = error_return
        self.error_call = error_call

    def __call__(self, function):
        def return_function(*args):
            try:
                return function(*args)
            except Exception as e:
                if type(e) not in self.errors:
                    raise e
                if self.error_call is not None:
                    self.error_call(e)
                return self.error_return

        return return_function


def call_error(e):
    pretty_json = json.dumps(json.loads(e.read()), indent=2, skipkeys=True)
    error = f'{e} {e.filename} : {pretty_json}'
    exceptions.append(error)


def call_artifact_error(e):
    artifact = e.filename.rsplit('/', 1)[1]
    print(tableTemplate.format(artifact, 'ERROR', f'{e}'))
    call_error(e)


def install_opener():
    opener = None
    if arguments.proxy and arguments.proxy_protocol:
        proxy_protocol = arguments.proxy_protocol.lower()

        if arguments.proxy_user and arguments.proxy_password:
            proxy_url = f'{proxy_protocol}://{arguments.proxy_user}:{arguments.proxy_password}@{arguments.proxy}'
        else:
            proxy_url = f'{proxy_protocol}://{arguments.proxy}'

        proxy_handler = urllib.request.ProxyHandler({'http': proxy_url, 'https': proxy_url})
        opener = urllib.request.build_opener(proxy_handler)
    else:
        opener = urllib.request.build_opener(urllib.request.HTTPSHandler())

    urllib.request.install_opener(opener)


@ErrorIgnore(errors=[HTTPError, Exception], error_return=None, error_call=call_error)
def get_download_links():
    req = urllib.request.Request(ENDPOINT + '/v1/data/reference-data/downloads?product=post,autocompletion,phone')
    req.add_header('Authorization', 'Bearer ' + arguments.api_key)

    with urllib.request.urlopen(req) as response:
        body = response.read()
        data = json.loads(body)

        if 'referenceData' not in data or 'installers' not in data:
            raise DownloadFailed("The list of available reference data could not be loaded. Please "
                                 "contact Uniserv support. Full body of response: ", data)

        return data


@ErrorIgnore(errors=[HTTPError, Exception], error_return=None, error_call=call_artifact_error)
def download_artifact(url):
    req = urllib.request.Request(url)
    req.add_header('Authorization', 'Bearer ' + arguments.api_key)

    with (urllib.request.urlopen(req) as response, open(response.info().get_filename(), 'wb') as out_file):
        shutil.copyfileobj(response, out_file)
        date = re.search("\d\d\d\d-\d\d-\d\d", url)
        file_name = response.info().get_filename()
        downloadedArtifacts.append(file_name)

        if date:
            version = date.group()
            print(tableTemplate.format(file_name, version[:7], response.headers["Last-Modified"]))
        else:
            print(tableTemplate.format(file_name, '', response.headers["Last-Modified"]))


def get_reference_artifact_name(item):
    if item['product'] == 'phone':
        return item['label'] + '.zip'
    return item['label'] + '_v' + item['format'] + '.zip'


def get_file_hash(name: str, algorithm: str) -> str:
    result = ''
    with open(name, 'rb', buffering=0) as file:
        if sys.version_info >= (3, 11):
            result = hashlib.file_digest(file, algorithm).hexdigest()
        else:
            hasher = hashlib.new(algorithm)
            chunk_size = 2 ** 20
            buffer = bytearray(chunk_size)
            while True:
                bytes_read = file.readinto(buffer)
                if bytes_read == 0:
                    break
                hasher.update(buffer[0:bytes_read])
            result = hasher.hexdigest()
    return result


@ErrorIgnore(errors=[HTTPError, Exception], error_return=None, error_call=None)
def should_download_artifact(item):
    artifact_to_check = get_reference_artifact_name(item)

    if Path(artifact_to_check).exists():
        with urllib.request.urlopen(item['hashUrl']) as response:
            repository_file_hash = response.read().decode('utf-8')
            existing_file_hash = get_file_hash(artifact_to_check, 'sha512')
            if secrets.compare_digest(repository_file_hash, existing_file_hash):
                skippedDownloads.append(
                    tableTemplate.format(artifact_to_check, item['version'][:7], response.headers["Last-Modified"]))
                return False

    return True


def download_version(installer_list):
    present_versions = [x for x in installer_list if x['label'] == 'version']

    if len(present_versions) == 1:
        download_artifact(present_versions[0]['url'])


def download_installer(installer_list):
    if arguments.installer is not None:
        chosen_installer = [x for x in installer_list if x['label'] == arguments.installer]
        for installer in chosen_installer:
            download_artifact(installer['url'])

    else:
        for installer in installer_list:
            if installer['label'] != 'version':
                download_artifact(installer['url'])


def download_granted_artifacts(file_list):
    print_table_header()

    for item in file_list['referenceData']:
        if should_download_artifact(item):
            download_artifact(item['url'])

    if downloadedArtifacts:
        download_version(file_list['installers'])
        download_installer(file_list['installers'])


def print_table_header():
    print(tableTemplate.format(*['Name', 'Generation date', 'Modified at']))


def evaluate_exceptions():
    if exceptions:
        raise DownloadFailed(exceptions)


def delete_existing_output_file():
    if os.path.exists(downloadOutputFileName):
        os.remove(downloadOutputFileName)


def write_output_file():
    if downloadedArtifacts:
        with open(downloadOutputFileName, "w") as file:
            file.write("\n".join(downloadedArtifacts))


def evaluate_skipped_downloads():
    if skippedDownloads:
        print("")
        print('Skipped download of following file(s), as they already have the newest version in the target folder:')
        print_table_header()

        for ignored in skippedDownloads:
            print(ignored)


if __name__ == '__main__':
    try:
        delete_existing_output_file()

        arguments = get_commandline_arguments()
        install_opener()

        print('Reference data download script \033[1m' + VERSION + '\033[0m')
        print('')

        download_links = get_download_links()

        # Could not get download links of permitted files
        if download_links is None:
            raise DownloadFailed(exceptions)

        download_granted_artifacts(download_links)

        evaluate_skipped_downloads()

        evaluate_exceptions()

        write_output_file()

        print("")
        print('Download of resources finished successfully.')

    except DownloadFailed as err:
        print("")
        print('Download of resources finished with following error(s):')
        for error in err.args[0]:
            print(error)
        exit(1)
